import ctypes
from typing import Dict, List, Optional, Union, Tuple

class ApiKeyResponse(ctypes.Structure):
    _fields_ = [("privateKey", ctypes.c_char_p), ("publicKey", ctypes.c_char_p), ("err", ctypes.c_char_p)]


class CreateOrderTxReq(ctypes.Structure):
    _fields_ = [
        ("MarketIndex", ctypes.c_uint8),
        ("ClientOrderIndex", ctypes.c_longlong),
        ("BaseAmount", ctypes.c_longlong),
        ("Price", ctypes.c_uint32),
        ("IsAsk", ctypes.c_uint8),
        ("Type", ctypes.c_uint8),
        ("TimeInForce", ctypes.c_uint8),
        ("ReduceOnly", ctypes.c_uint8),
        ("TriggerPrice", ctypes.c_uint32),
        ("OrderExpiry", ctypes.c_longlong),
    ]


class StrOrErr(ctypes.Structure):
    _fields_ = [("str", ctypes.c_char_p), ("err", ctypes.c_char_p)]


class SignedTxResponse(ctypes.Structure):
    _fields_ = [
        ("txType", ctypes.c_uint8),
        ("txInfo", ctypes.c_char_p),
        ("txHash", ctypes.c_char_p),
        ("messageToSign", ctypes.c_char_p),
        ("err", ctypes.c_char_p),
    ]

lighterSigner = None

def load_lighter_library(path):
    global lighterSigner

    if lighterSigner is not None:
        return lighterSigner

    lighterSigner = ctypes.CDLL(path)
    lighterSigner.GenerateAPIKey.argtypes = [ctypes.c_char_p]
    lighterSigner.GenerateAPIKey.restype = ApiKeyResponse

    lighterSigner.CreateClient.argtypes = [ctypes.c_char_p, ctypes.c_char_p, ctypes.c_int, ctypes.c_int, ctypes.c_longlong]
    lighterSigner.CreateClient.restype = ctypes.c_char_p

    lighterSigner.CheckClient.argtypes = [ctypes.c_int, ctypes.c_longlong]
    lighterSigner.CheckClient.restype = ctypes.c_char_p

    lighterSigner.SignChangePubKey.argtypes = [ctypes.c_char_p, ctypes.c_longlong, ctypes.c_int, ctypes.c_longlong]
    lighterSigner.SignChangePubKey.restype = SignedTxResponse

    lighterSigner.SignCreateOrder.argtypes = [ctypes.c_int, ctypes.c_longlong, ctypes.c_longlong, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int,
                                            ctypes.c_int, ctypes.c_int, ctypes.c_longlong, ctypes.c_longlong, ctypes.c_int, ctypes.c_longlong]
    lighterSigner.SignCreateOrder.restype = SignedTxResponse

    lighterSigner.SignCreateGroupedOrders.argtypes = [ctypes.c_uint8, ctypes.POINTER(CreateOrderTxReq), ctypes.c_int, ctypes.c_longlong, ctypes.c_int, ctypes.c_longlong]
    lighterSigner.SignCreateGroupedOrders.restype = SignedTxResponse

    lighterSigner.SignCancelOrder.argtypes = [ctypes.c_int, ctypes.c_longlong, ctypes.c_longlong, ctypes.c_int, ctypes.c_longlong]
    lighterSigner.SignCancelOrder.restype = SignedTxResponse

    lighterSigner.SignWithdraw.argtypes = [ctypes.c_int, ctypes.c_int, ctypes.c_longlong, ctypes.c_longlong, ctypes.c_int, ctypes.c_longlong]
    lighterSigner.SignWithdraw.restype = SignedTxResponse

    lighterSigner.SignCreateSubAccount.argtypes = [ctypes.c_longlong, ctypes.c_int, ctypes.c_longlong]
    lighterSigner.SignCreateSubAccount.restype = SignedTxResponse

    lighterSigner.SignCancelAllOrders.argtypes = [ctypes.c_int, ctypes.c_longlong, ctypes.c_longlong, ctypes.c_int, ctypes.c_longlong]
    lighterSigner.SignCancelAllOrders.restype = SignedTxResponse

    lighterSigner.SignModifyOrder.argtypes = [ctypes.c_int, ctypes.c_longlong, ctypes.c_longlong, ctypes.c_longlong, ctypes.c_longlong, ctypes.c_longlong, ctypes.c_int, ctypes.c_longlong]
    lighterSigner.SignModifyOrder.restype = SignedTxResponse

    lighterSigner.SignTransfer.argtypes = [ctypes.c_longlong, ctypes.c_int16, ctypes.c_int8, ctypes.c_int8, ctypes.c_longlong, ctypes.c_longlong, ctypes.c_char_p, ctypes.c_longlong, ctypes.c_int, ctypes.c_longlong]
    lighterSigner.SignTransfer.restype = SignedTxResponse

    lighterSigner.SignCreatePublicPool.argtypes = [ctypes.c_longlong, ctypes.c_int, ctypes.c_longlong, ctypes.c_longlong, ctypes.c_int, ctypes.c_longlong]
    lighterSigner.SignCreatePublicPool.restype = SignedTxResponse

    lighterSigner.SignUpdatePublicPool.argtypes = [ctypes.c_longlong, ctypes.c_int, ctypes.c_longlong, ctypes.c_int, ctypes.c_longlong, ctypes.c_int, ctypes.c_longlong]
    lighterSigner.SignUpdatePublicPool.restype = SignedTxResponse

    lighterSigner.SignMintShares.argtypes = [ctypes.c_longlong, ctypes.c_longlong, ctypes.c_longlong, ctypes.c_int, ctypes.c_longlong]
    lighterSigner.SignMintShares.restype = SignedTxResponse

    lighterSigner.SignBurnShares.argtypes = [ctypes.c_longlong, ctypes.c_longlong, ctypes.c_longlong, ctypes.c_int, ctypes.c_longlong]
    lighterSigner.SignBurnShares.restype = SignedTxResponse

    lighterSigner.SignUpdateLeverage.argtypes = [ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_longlong, ctypes.c_int, ctypes.c_longlong]
    lighterSigner.SignUpdateLeverage.restype = SignedTxResponse

    lighterSigner.CreateAuthToken.argtypes = [ctypes.c_longlong, ctypes.c_int, ctypes.c_longlong]
    lighterSigner.CreateAuthToken.restype = StrOrErr

    # Note: SwitchAPIKey is no longer exported in the new binary
    # All functions now take api_key_index directly, so switching is handled via parameters

    lighterSigner.SignUpdateMargin.argtypes = [ctypes.c_int, ctypes.c_longlong, ctypes.c_int, ctypes.c_longlong, ctypes.c_int, ctypes.c_longlong]
    lighterSigner.SignUpdateMargin.restype = SignedTxResponse
    return lighterSigner

def decode_tx_info(result: SignedTxResponse) -> Union[Tuple[str, str, str, None], Tuple[None, None, None, str]]:
    if result.err:
        error = result.err.decode("utf-8")
        return None, None, None, error
    
    # Use txType from response if available, otherwise use the provided type
    tx_type = result.txType
    tx_info_str = result.txInfo.decode("utf-8") if result.txInfo else None
    tx_hash_str = result.txHash.decode("utf-8") if result.txHash else None

    return tx_type, tx_info_str, tx_hash_str, None

def decode_auth(result: StrOrErr) -> Union[Tuple[str, None], Tuple[None, str]]:
    if result.err:
        error = result.err.decode("utf-8")
        return None, error

    token = result.str.decode('utf-8') if result.str else None
    return token, None



# Copyright (c) 2022 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for SHA-3 algorithms."""

# Imports
import hashlib
from typing import Union

from Crypto.Hash import SHA3_256, keccak

from ..misc import AlgoUtils


HASHLIB_USE_SHA3_256: bool = "sha3_256" in hashlib.algorithms_available


class Kekkak256:
    """
    Kekkak-256 class.
    It computes digests using Kekkak-256 algorithm.
    """

    @staticmethod
    def QuickDigest(data: Union[bytes, str]) -> bytes:
        """
        Compute the digest (quick version).

        Args:
            data (str or bytes): Data

        Returns:
            bytes: Computed digest
        """
        return keccak.new(data=AlgoUtils.Encode(data), digest_bits=256).digest()

    @staticmethod
    def DigestSize() -> int:
        """
        Get the digest size in bytes.

        Returns:
            int: Digest size in bytes
        """
        return (hashlib.new("sha3_256").digest_size
                if HASHLIB_USE_SHA3_256
                else keccak.new(digest_bits=256).digest_size)


class Sha3_256:     # noqa: N801
    """
    SHA3-256 class.
    It computes digests using SHA3-256 algorithm.
    """

    @staticmethod
    def QuickDigest(data: Union[bytes, str]) -> bytes:
        """
        Compute the digest (quick version).

        Args:
            data (str or bytes): Data

        Returns:
            bytes: Computed digest
        """
        if HASHLIB_USE_SHA3_256:
            return hashlib.new("sha3_256", AlgoUtils.Encode(data)).digest()
        # Use Cryptodome if not implemented in hashlib
        return SHA3_256.new(AlgoUtils.Encode(data)).digest()

    @staticmethod
    def DigestSize() -> int:
        """
        Get the digest size in bytes.

        Returns:
            int: Digest size in bytes
        """
        return (hashlib.new("sha3_256").digest_size
                if HASHLIB_USE_SHA3_256
                else SHA3_256.digest_size)



import asyncio

# Test by running:
# - python python/ccxt/pro/test/base/test_close.py
# - python python/ccxt/pro/test/base/test_future.py
class Future(asyncio.Future):

    def resolve(self, result=None):
        if not self.done():
            self.set_result(result)

    def reject(self, error=None):
        if not self.done():
            self.set_exception(error)

    @classmethod
    def race(cls, futures):
        """
        Return a Future that resolves/rejects with the first completed input future.

        IMPORTANT:
        - No asyncio.create_task(asyncio.wait(...)) => avoids massive Task churn.
        - We attach done callbacks and detach them immediately once a winner is chosen.
        """

        out = cls()

        if not futures:
            out.set_exception(Exception("Future.race() called with empty futures"))
            return out

        callbacks = {}  # future -> callback

        def detach_all():
            for f, cb in list(callbacks.items()):
                try:
                    f.remove_done_callback(cb)
                except Exception:
                    pass
            callbacks.clear()

        def settle_from(f):
            if out.done():
                detach_all()
                return
            try:
                if f.cancelled():
                    out.cancel()
                    return
                err = f.exception()
                if err is not None:
                    out.set_exception(err)
                else:
                    out.set_result(f.result())
            finally:
                detach_all()

        # Fast path: if any future is already done, settle immediately.
        for f in futures:
            if f.done():
                settle_from(f)
                return out

        # Attach callbacks to settle on first completion.
        for f in futures:
            def _cb(ff, _settle=settle_from):
                _settle(ff)
            callbacks[f] = _cb
            f.add_done_callback(_cb)

        # If the returned future is cancelled externally, detach callbacks.
        def _out_done(_):
            detach_all()
        out.add_done_callback(_out_done)

        return out
